﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2010 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Analysis.FitnessLandscape {

  [StorableClass]
  public class InformationAnalyzer : SingleSuccessorOperator, IQualityTrailAnalyzer {
    public bool EnabledByDefault {
      get { return false; }
    }

    #region Parameters
    public LookupParameter<DataTable> QualityTrailParameter {
      get { return (LookupParameter<DataTable>)Parameters["Quality Trail"]; }
    }
    public LookupParameter<ResultCollection> ResultsParameter {
      get { return (LookupParameter<ResultCollection>)Parameters["Results"]; }
    }
    public ValueLookupParameter<IntValue> NQuantilesParameter {
      get { return (ValueLookupParameter<IntValue>)Parameters["NQuantiles"]; }
    }
    public ValueLookupParameter<IntValue> ShapeSizeParameter {
      get { return (ValueLookupParameter<IntValue>)Parameters["ShapeSize"]; }
    }
    public LookupParameter<DoubleValue> InformationContentParameter {
      get { return (LookupParameter<DoubleValue>)Parameters["InformationContent"]; }
    }
    public LookupParameter<DoubleValue> PartialInformationContentParameter {
      get { return (LookupParameter<DoubleValue>)Parameters["PartialInformationContent"]; }
    }
    public LookupParameter<DoubleValue> DensityBasinInformationParameter {
      get { return (LookupParameter<DoubleValue>)Parameters["DensityBasinInformation"]; }
    }
    public LookupParameter<DoubleValue> TotalEntropyParameter {
      get { return (LookupParameter<DoubleValue>)Parameters["TotalEntropy"]; }
    }
    public LookupParameter<DoubleValue> InformationStabilityParameter {
      get { return (LookupParameter<DoubleValue>)Parameters["InformationStability"]; }
    }
    public LookupParameter<DoubleValue> RegularityParameter {
      get { return (LookupParameter<DoubleValue>)Parameters["Regularity"]; }
    }
    public LookupParameter<DoubleValue> DiversityParameter {
      get { return (LookupParameter<DoubleValue>)Parameters["Diversity"]; }
    }
    public LookupParameter<DoubleValue> PeakInformationContentParameter {
      get { return (LookupParameter<DoubleValue>)Parameters["PeakInformationContent"]; }
    }
    public LookupParameter<DoubleValue> PeakDensityBasinInformationParameter {
      get { return (LookupParameter<DoubleValue>)Parameters["PeakDensityBasinInformation"]; }
    }
    public ILookupParameter<DataTable> InformationAnalysisParameter {
      get { return (ILookupParameter<DataTable>)Parameters["InformationAnalysis"]; }
    }
    #endregion

    [StorableConstructor]
    protected InformationAnalyzer(bool deserializing) : base(deserializing) { }
    protected InformationAnalyzer(InformationAnalyzer original, Cloner cloner) : base(original, cloner) { }

    public InformationAnalyzer() {
      Parameters.Add(new LookupParameter<DataTable>("Quality Trail", "The qualities of the solutions"));
      Parameters.Add(new LookupParameter<ResultCollection>("Results", "The collection of all results of this algorithm"));

      Parameters.Add(new ValueLookupParameter<IntValue>("NQuantiles", "The number of delta quantiles to display", new IntValue(20)));

      Parameters.Add(new LookupParameter<DoubleValue>("InformationContent", "The information content H(0) at eps = 0"));
      Parameters.Add(new LookupParameter<DoubleValue>("PartialInformationContent", "Partial information content M(0) at eps = 0"));
      Parameters.Add(new LookupParameter<DoubleValue>("DensityBasinInformation", "Density Basin Information h(0) at eps = 0"));
      Parameters.Add(new LookupParameter<DoubleValue>("TotalEntropy", "The overall disorder in the trajectory"));
      Parameters.Add(new LookupParameter<DoubleValue>("InformationStability", "Information Stability Value"));
      Parameters.Add(new LookupParameter<DoubleValue>("Regularity", "The number of different quality differences"));
      Parameters.Add(new LookupParameter<DoubleValue>("Diversity", "The number of different quality values"));
      Parameters.Add(new LookupParameter<DoubleValue>("PeakInformationContent", "Maximum information content at any quality delta."));
      Parameters.Add(new LookupParameter<DoubleValue>("PeakDensityBasinInformation", "Maximum density basin information at any quality delta."));

      Parameters.Add(new LookupParameter<DataTable>("InformationAnalysis", "Graphical analysis of information theoretic features."));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new InformationAnalyzer(this, cloner);
    }

    public override IOperation Apply() {
      var qualityTrail = QualityTrailParameter.ActualValue;
      if (qualityTrail == null || qualityTrail.Rows.Count == 0) return base.Apply();

      var qualities = QualityTrailParameter.ActualValue.Rows.First().Values.ToList();
      if (qualities.Count <= 2) return base.Apply();

      var nQuantiles = NQuantilesParameter.ActualValue.Value;
      var results = ResultsParameter.ActualValue;

      var analysis = new InformationAnalysis(qualities, nQuantiles);
      var ic = new DoubleValue(analysis.InformationContent.FirstOrDefault());
      InformationContentParameter.ActualValue = ic;
      results.AddOrUpdateResult(InformationContentParameter.Name, ic);
      var pic = new DoubleValue(analysis.PartialInformationContent.FirstOrDefault());
      PartialInformationContentParameter.ActualValue = pic;
      results.AddOrUpdateResult(PartialInformationContentParameter.Name, pic);
      var dbi = new DoubleValue(analysis.DensityBasinInformation.FirstOrDefault());
      DensityBasinInformationParameter.ActualValue = dbi;
      results.AddOrUpdateResult(DensityBasinInformationParameter.Name, dbi);
      var @is = new DoubleValue(analysis.InformationStability);
      InformationStabilityParameter.ActualValue = @is;
      results.AddOrUpdateResult(InformationStabilityParameter.Name, @is);
      var regularity = new DoubleValue(1.0 * analysis.Regularity / qualities.Count);
      RegularityParameter.ActualValue = regularity;
      results.AddOrUpdateResult(RegularityParameter.Name, regularity);
      var diversity = new DoubleValue(1.0 * analysis.Diversity / qualities.Count);
      DiversityParameter.ActualValue = diversity;
      results.AddOrUpdateResult(DiversityParameter.Name, diversity);
      var totalEntropy = new DoubleValue(analysis.TotalEntropy.FirstOrDefault());
      TotalEntropyParameter.ActualValue = totalEntropy;
      results.AddOrUpdateResult(TotalEntropyParameter.Name, totalEntropy);
      var peakIc = new DoubleValue(analysis.PeakInformationContent.Value);
      PeakInformationContentParameter.ActualValue = peakIc;
      results.AddOrUpdateResult(PeakInformationContentParameter.Name, peakIc);
      var peakDbi = new DoubleValue(analysis.PeakDensityBasinInformation.Value);
      PeakDensityBasinInformationParameter.ActualValue = peakDbi;
      results.AddOrUpdateResult(PeakDensityBasinInformationParameter.Name, peakDbi);

      var itable = GetInformationTable(analysis);
      results.AddOrUpdateResult(InformationAnalysisParameter.Name, itable);
      return base.Apply();
    }

    private static DataTable GetInformationTable(InformationAnalysis analysis) {
      var dt = new DataTable("Information Analysis");
      var ic = new DataRow("Information Content");
      var pic = new DataRow("Partial Information Content");
      var dbi = new DataRow("Density Basin Information");
      var te = new DataRow("Total Entropy");
      var e = new DataRow("Epsilon");
      e.VisualProperties.SecondYAxis = true;
      ic.Values.AddRange(analysis.InformationContent);
      pic.Values.AddRange(analysis.PartialInformationContent);
      dbi.Values.AddRange(analysis.DensityBasinInformation);
      te.Values.AddRange(analysis.TotalEntropy);
      e.Values.AddRange(analysis.QualityDelta);
      dt.Rows.Add(ic);
      dt.Rows.Add(pic);
      dt.Rows.Add(dbi);
      dt.Rows.Add(te);
      dt.Rows.Add(e);
      return dt;
    }
  }
}
